"""Test trace estimation."""

from matfree import stochtrace
from matfree.backend import func, linalg, np, prng, tree


def test_trace():
    """Assert that traces are estimated correctly."""

    def fun(x):
        """Create a nonlinear, to-be-differentiated function."""
        fx = np.sin(np.flip(np.cos(x["params"])) + 1.0) * np.sin(x["params"])
        return {"params": fx}

    key = prng.prng_key(seed=2)

    # Linearise function
    x0 = prng.uniform(key, shape=(4,))  # random lin. point
    args_like = {"params": x0}
    _, jvp = func.linearize(fun, args_like)
    J = func.jacfwd(fun)(args_like)["params"]["params"]
    expected = linalg.trace(J)

    # Estimate the matrix function
    problem = stochtrace.integrand_trace()
    sampler = stochtrace.sampler_normal(args_like, num=100_000)
    estimate = stochtrace.estimator(problem, sampler=sampler)
    received = estimate(jvp, key)

    def compare(a, b):
        return np.allclose(a, b, rtol=1e-2)

    assert tree.tree_all(tree.tree_map(compare, received, expected))



"""Test the adjoint of tri-diagonalisation."""

from matfree import decomp, test_util
from matfree.backend import func, linalg, np, prng, testing, tree


@testing.parametrize("reortho", ["full", "none"])
def test_adjoint_vjp_matches_jax_vjp(reortho, n=10, krylov_num_matvecs=4):
    """Test that the custom VJP yields the same output as autodiff."""
    # Set up a test-matrix
    eigvals = prng.uniform(prng.prng_key(2), shape=(n,)) + 1.0
    matrix = test_util.symmetric_matrix_from_eigenvalues(eigvals)
    params = _sym(matrix)

    def matvec(s, p):
        return (p + p.T) @ s

    # Set up an initial vector
    vector = prng.normal(prng.prng_key(1), shape=(n,))

    # Flatten the inputs
    flat, unflatten = tree.ravel_pytree((vector, params))

    # Construct a vector-to-vector decomposition function
    def decompose(f, *, custom_vjp):
        kwargs = {"reortho": reortho, "custom_vjp": custom_vjp, "materialize": False}
        algorithm = decomp.tridiag_sym(krylov_num_matvecs, **kwargs)
        output = algorithm(matvec, *unflatten(f))
        return tree.ravel_pytree(output)[0]

    # Construct the two implementations
    reference = func.jit(func.partial(decompose, custom_vjp=False))
    implementation = func.jit(func.partial(decompose, custom_vjp=True))

    # Compute both VJPs
    fx_ref, vjp_ref = func.vjp(reference, flat)
    fx_imp, vjp_imp = func.vjp(implementation, flat)
    # Assert that the forward-passes are identical
    assert np.allclose(fx_ref, fx_imp)

    # Assert that the VJPs into a bunch of random directions are identical
    for seed in [4, 5, 6]:
        key = prng.prng_key(seed)
        dnu = prng.normal(key, shape=np.shape(reference(flat)))
        assert np.allclose(*vjp_ref(dnu), *vjp_imp(dnu), atol=1e-4, rtol=1e-4)


def _sym(m):
    return np.triu(m) - linalg.diagonal_matrix(0.5 * linalg.diagonal(m))



"""Test-utilities."""

import pytest
import pytest_cases


def fixture(name=None):
    return pytest_cases.fixture(name=name)


def parametrize(argnames, argvalues, /):
    return pytest.mark.parametrize(argnames, argvalues)


def parametrize_with_cases(argnames, /, cases, prefix):
    return pytest_cases.parametrize_with_cases(argnames, cases=cases, prefix=prefix)


def raises(err, /, match):
    return pytest.raises(err, match=match)


def warns(warning, /):
    return pytest.warns(warning)



"""Compute log-determinants with stochastic Lanczos quadrature.

Log-determinant estimation can be implemented with stochastic Lanczos quadrature,
which can be loosely interpreted as an extension of Hutchinson's trace estimator.
"""

import jax
import jax.numpy as jnp

from matfree import decomp, funm, stochtrace

# Set up a matrix.

nhidden, nrows = 6, 5
A = jnp.reshape(jnp.arange(1.0, 1.0 + nhidden * nrows), (nhidden, nrows))
A /= nhidden * nrows


def matvec(x):
    """Compute a matrix-vector product."""
    return A.T @ (A @ x) + x


x_like = jnp.ones((nrows,), dtype=float)  # use to determine shapes of vectors


# Estimate log-determinants with stochastic Lanczos quadrature.

num_matvecs = 3
tridiag_sym = decomp.tridiag_sym(num_matvecs)
problem = funm.integrand_funm_sym_logdet(tridiag_sym)
sampler = stochtrace.sampler_normal(x_like, num=1_000)
estimator = stochtrace.estimator(problem, sampler=sampler)
logdet = estimator(matvec, jax.random.PRNGKey(1))
print(logdet)

# For comparison:

print(jnp.linalg.slogdet(A.T @ A + jnp.eye(nrows)))


# We can compute the log determinant of a matrix
# of the form $M = B^\top B$, purely based
# on arithmetic with $B$; no need to assemble $M$:

A = jnp.reshape(jnp.arange(1.0, 1.0 + nrows**2), (nrows, nrows))
A += jnp.eye(nrows)
A /= nrows**2


def matvec_half(x):
    """Compute a matrix-vector product."""
    return A @ x


num_matvecs = 3
bidiag = decomp.bidiag(num_matvecs)
problem = funm.integrand_funm_product_logdet(bidiag)
sampler = stochtrace.sampler_normal(x_like, num=1_000)
estimator = stochtrace.estimator(problem, sampler=sampler)
logdet = estimator(matvec_half, jax.random.PRNGKey(1))
print(logdet)

# Internally, Matfree uses JAX's vector-Jacobian products to
# turn the matrix-vector product into a vector-matrix product.
#
# For comparison:

print(jnp.linalg.slogdet(A.T @ A))



"""Test preconditioning with partial Cholesky decompositions."""

from matfree import low_rank, test_util
from matfree.backend import linalg, np


def test_preconditioner_solves_correctly(n=10):
    # Create a relatively ill-conditioned matrix
    cov_eig = 1.5 ** np.arange(-n // 2, n // 2, step=1.0)
    cov = test_util.symmetric_matrix_from_eigenvalues(cov_eig)

    def element(i, j):
        return cov[i, j]

    # Assert that the Cholesky decomposition is full-rank.
    # This is important to ensure that the test below makes sense.
    cholesky = low_rank.cholesky_partial(element, nrows=n, rank=n)
    matrix, _info = cholesky()
    assert np.allclose(matrix @ matrix.T, cov)

    # Set up the test-problem
    small_value = 1e-1
    b = np.arange(1.0, 1 + len(cov))
    b /= linalg.vector_norm(b)

    # Solve the linear system
    cov_added = cov + small_value * np.eye(len(cov))
    expected = linalg.solve(cov_added, b)

    # Derive the preconditioner
    precondition = low_rank.preconditioner(cholesky)
    received, info = precondition(b, small_value)

    # Test that the preconditioner solves correctly
    assert np.allclose(received, expected, rtol=1e-2)



"""Matrix-free matrix decompositions.

This module includes various Lanczos-decompositions of matrices
(tri-diagonal, bi-diagonal, etc.).

For stochastic Lanczos quadrature and
matrix-function-vector products, see
[matfree.funm][matfree.funm].
"""

from matfree.backend import containers, control_flow, func, linalg, np, tree
from matfree.backend.typing import Array, Callable, Union


class _DecompResult(containers.NamedTuple):
    # If an algorithm returns a single Q, place it here.
    # If it returns multiple Qs, stack them
    # into a tuple and place them here.
    Q_tall: Union[Array, tuple[Array, ...]]

    # If an algorithm returns a materialized matrix,
    # place it here. If it returns a sparse representation
    # (e.g. two vectors representing diagonals), place it here
    J_small: Union[Array, tuple[Array, ...]]

    residual: Array
    init_length_inv: Array


def tridiag_sym(
    num_matvecs: int,
    /,
    *,
    materialize: bool = True,
    reortho: str = "full",
    custom_vjp: bool = True,
):
    r"""Construct an implementation of **tridiagonalisation**.

    Uses pre-allocation, and full reorthogonalisation if `reortho` is set to `"full"`.
    It tends to be a good idea to use full reorthogonalisation.

    This algorithm assumes a **symmetric matrix**.

    Decompose a matrix into a product of orthogonal-**tridiagonal**-orthogonal matrices.
    Use this algorithm for approximate **eigenvalue** decompositions.

    Setting `custom_vjp` to `True` implies using efficient, numerically stable
    gradients of the Lanczos iteration according to what has been proposed by
    Krämer et al. (2024).
    These gradients are exact, so there is little reason not to use them.
    If you use this configuration, please consider
    citing Krämer et al. (2024; bibtex below).

    ??? note "BibTex for Krämer et al. (2024)"
        ```bibtex
        @article{kraemer2024gradients,
            title={Gradients of functions of large matrices},
            author={Kr\"amer, Nicholas and Moreno-Mu\~noz, Pablo and
            Roy, Hrittik and Hauberg, S{\o}ren},
            journal={arXiv preprint arXiv:2405.17277},
            year={2024}
        }
        ```

    Parameters
    ----------
    num_matvecs
        The number of matrix-vector products aka the depth of the Krylov space.
        The deeper the Krylov space, the more accurate the factorisation tends to be.
        However, the computational complexity increases linearly
        with the number of matrix-vector products.
    materialize
        The value of this flag indicates whether the tridiagonal matrix
        should be returned in a sparse format (which means, as a tuple of diagonas)
        or as a dense matrix.
        The dense matrix is helpful if different decompositions should be used
        interchangeably. The sparse representation requires less memory.
    reortho
        The value of this parameter indicates whether to reorthogonalise the
        basis vectors during the forward pass.
        Reorthogonalisation makes the forward pass more expensive, but helps
        (significantly) with numerical stability.
    custom_vjp
        The value of this flag indicates whether to use a custom vector-Jacobian
        product as proposed by Krämer et al. (2024; bibtex above).
        Generally, using a custom VJP tends to be a good idea.
        However, due to JAX's mechanics, a custom VJP precludes the use of forward-mode
        differentiation
        ([see here](https://jax.readthedocs.io/en/latest/_autosummary/jax.custom_vjp.html)),
        so don't use a custom VJP if you need forward-mode differentiation.

    Returns
    -------
    decompose
        A decomposition function that maps
        ``(matvec, vector, *params)`` to the decomposition.
        The decomposition is a tuple of (nested) arrays.
        The first element is the Krylov basis,
        the second element represents the tridiagonal matrix
        (how it is represented depends on the value of ``materialize''),
        the third element is
        the residual, and the fourth element is
        the (inverse of the) length of the initial vector.
    """
    if reortho == "full":
        return _tridiag_reortho_full(
            num_matvecs, custom_vjp=custom_vjp, materialize=materialize
        )
    if reortho == "none":
        return _tridiag_reortho_none(
            num_matvecs, custom_vjp=custom_vjp, materialize=materialize
        )

    msg = f"reortho={reortho} unsupported. Choose eiter {'full', 'none'}."
    raise ValueError(msg)


def _tridiag_reortho_full(num_matvecs: int, /, *, custom_vjp: bool, materialize: bool):
    # Implement via Arnoldi to use the reorthogonalised adjoints.
    # The complexity difference is minimal with full reortho.
    alg = hessenberg(num_matvecs, custom_vjp=custom_vjp, reortho="full")

    def estimate(matvec, vec, *params):
        Q, H, v, norm = alg(matvec, vec, *params)

        T = 0.5 * (H + H.T)
        diags = linalg.diagonal(T, offset=0)
        offdiags = linalg.diagonal(T, offset=1)

        matrix = (diags, offdiags)
        if materialize:
            matrix = _todense_tridiag_sym(diags, offdiags)

        return _DecompResult(
            Q_tall=Q, J_small=matrix, residual=v, init_length_inv=1.0 / norm
        )

    return estimate


def _todense_tridiag_sym(diag, off_diag):
    diag = linalg.diagonal_matrix(diag)
    offdiag1 = linalg.diagonal_matrix(off_diag, -1)
    offdiag2 = linalg.diagonal_matrix(off_diag, 1)
    return diag + offdiag1 + offdiag2


def _tridiag_reortho_none(num_matvecs: int, /, *, custom_vjp: bool, materialize: bool):
    def estimate(matvec, vec, *params):
        if num_matvecs < 0 or num_matvecs > len(vec):
            msg = _error_num_matvecs(num_matvecs, maxval=len(vec), minval=0)
            raise ValueError(msg)

        (Q, H), (q, b) = _estimate(matvec, vec, *params)
        v = b * q

        if materialize:
            H = _todense_tridiag_sym(*H)

        length = linalg.vector_norm(vec)
        return _DecompResult(
            Q_tall=Q.T, J_small=H, residual=v, init_length_inv=1.0 / length
        )

    def _estimate(matvec, vec, *params):
        *values, _ = _tridiag_forward(matvec, num_matvecs, vec, *params)
        return values

    def estimate_fwd(matvec, vec, *params):
        value = _estimate(matvec, vec, *params)
        return value, (value, (linalg.vector_norm(vec), *params))

    def estimate_bwd(matvec, cache, vjp_incoming):
        # Read incoming gradients and stack related quantities
        (dxs, (dalphas, dbetas)), (dx_last, dbeta_last) = vjp_incoming
        dxs = np.concatenate((dxs, dx_last[None]))
        dbetas = np.concatenate((dbetas, dbeta_last[None]))

        # Read the cache and stack related quantities
        ((xs, (alphas, betas)), (x_last, beta_last)), (vector_norm, *params) = cache
        xs = np.concatenate((xs, x_last[None]))
        betas = np.concatenate((betas, beta_last[None]))

        # Compute the adjoints, discard the adjoint states, and return the gradients
        grads, _lambdas_and_mus_and_nus = _tridiag_adjoint(
            matvec=matvec,
            params=params,
            initvec_norm=vector_norm,
            alphas=alphas,
            betas=betas,
            xs=xs,
            dalphas=dalphas,
            dbetas=dbetas,
            dxs=dxs,
        )
        return grads

    if custom_vjp:
        _estimate = func.custom_vjp(_estimate, nondiff_argnums=[0])
        _estimate.defvjp(estimate_fwd, estimate_bwd)  # type: ignore

    return estimate


def _tridiag_forward(matvec, num_matvecs, vec, *params):
    # Pre-allocate
    vectors = np.zeros((num_matvecs + 1, len(vec)))
    offdiags = np.zeros((num_matvecs,))
    diags = np.zeros((num_matvecs,))

    # Normalize (not all Lanczos implementations do that)
    v0 = vec / linalg.vector_norm(vec)
    vectors = vectors.at[0].set(v0)

    # Lanczos initialisation
    ((v1, offdiag), diag) = _tridiag_fwd_init(matvec, v0, *params)

    if num_matvecs == 0:
        decomposition = vectors[:-1], (diags, offdiags[:-1])
        remainder = v1, offdiag
        return decomposition, remainder, 1 / linalg.vector_norm(vec)

    # Store results
    k = 0
    vectors = vectors.at[k + 1].set(v1)
    offdiags = offdiags.at[k].set(offdiag)
    diags = diags.at[k].set(diag)

    # Run Lanczos-loop
    init = (v1, offdiag, v0), (vectors, diags, offdiags)
    step_fun = func.partial(_tridiag_fwd_step, matvec, params)
    (v1, offdiag, _), (vectors, diags, offdiags) = control_flow.fori_loop(
        lower=1, upper=num_matvecs, body_fun=step_fun, init_val=init
    )

    # Reorganise the outputs
    decomposition = vectors[:-1], (diags, offdiags[:-1])
    remainder = v1, offdiag
    return decomposition, remainder, 1 / linalg.vector_norm(vec)


def _tridiag_fwd_init(matvec, vec, *params):
    """Initialize Lanczos' algorithm.

    Solve A x_{k} = a_k x_k + b_k x_{k+1}
    for x_{k+1}, a_k, and b_k, using
    orthogonality of the x_k.
    """
    a = vec @ (matvec(vec, *params))
    r = (matvec(vec, *params)) - a * vec
    b = linalg.vector_norm(r)
    x = r / b
    return (x, b), a


def _tridiag_fwd_step(matvec, params, i, val):
    (v1, offdiag, v0), (vectors, diags, offdiags) = val
    ((v1, offdiag), diag), v0 = (
        _tridiag_fwd_step_apply(matvec, v1, offdiag, v0, *params),
        v1,
    )

    # Store results
    vectors = vectors.at[i + 1].set(v1)
    offdiags = offdiags.at[i].set(offdiag)
    diags = diags.at[i].set(diag)

    return (v1, offdiag, v0), (vectors, diags, offdiags)


def _tridiag_fwd_step_apply(matvec, vec, b, vec_previous, *params):
    """Apply Lanczos' recurrence."""
    a = vec @ (matvec(vec, *params))
    r = matvec(vec, *params) - a * vec - b * vec_previous
    b = linalg.vector_norm(r)
    x = r / b
    return (x, b), a


def _tridiag_adjoint(
    *, matvec, params, initvec_norm, alphas, betas, xs, dalphas, dbetas, dxs
):
    def adjoint_step(xi_and_lambda, inputs):
        return _tridiag_adjoint_step(
            *xi_and_lambda, matvec=matvec, params=params, **inputs
        )

    # Scan over all input gradients and output values
    xs0 = xs
    xs0 = xs0.at[-1, :].set(np.zeros_like(xs[-1, :]))

    loop_over = {
        "dx": dxs[:-1],
        "da": dalphas,
        "db": dbetas,
        "xs": (xs[1:], xs[:-1]),
        "a": alphas,
        "b": betas,
    }
    init_val = (xs0, -dxs[-1], np.zeros_like(dxs[-1]))
    (_, lambda_1, _lambda_2), (grad_summands, *other) = control_flow.scan(
        adjoint_step, init_val, xs=loop_over, reverse=True
    )

    # Compute the gradients
    grad_matvec = tree.tree_map(lambda s: np.sum(s, axis=0), grad_summands)
    grad_initvec = ((lambda_1.T @ xs[0]) * xs[0] - lambda_1) / initvec_norm

    # Return values
    return (grad_initvec, grad_matvec), (lambda_1, *other)


def _tridiag_adjoint_step(
    xs_all, xi, lambda_plus, /, *, matvec, params, dx, da, db, xs, a, b
):
    # Read inputs
    (xplus, x) = xs

    # Apply formula
    xi /= b
    mu = db - lambda_plus.T @ x + xplus.T @ xi
    nu = da + x.T @ xi
    lambda_ = -xi + mu * xplus + nu * x

    # Value-and-grad of matrix-vector product
    matvec_lambda, vjp = func.vjp(lambda *p: matvec(lambda_, *p), *params)
    (gradient_increment,) = vjp(x)

    # Prepare next step
    xi = -dx - matvec_lambda + a * lambda_ + b * lambda_plus - b * nu * xplus

    # Return values
    return (xs_all, xi, lambda_), (gradient_increment, lambda_, mu, nu, xi)


def hessenberg(
    num_matvecs, /, *, reortho: str, custom_vjp: bool = True, reortho_vjp: str = "match"
):
    r"""Construct a **Hessenberg-factorisation** via the Arnoldi iteration.

    Uses pre-allocation, and full reorthogonalisation if `reortho` is set to `"full"`.
    It tends to be a good idea to use full reorthogonalisation.

    This algorithm works for **arbitrary matrices**.

    Setting `custom_vjp` to `True` implies using efficient, numerically stable
    gradients of the Arnoldi iteration according to what has been proposed by
    Krämer et al. (2024).
    These gradients are exact, so there is little reason not to use them.
    If you use this configuration,
    please consider citing Krämer et al. (2024; bibtex below).

    ??? note "BibTex for Krämer et al. (2024)"
        ```bibtex
        @article{kraemer2024gradients,
            title={Gradients of functions of large matrices},
            author={Kr\"amer, Nicholas and Moreno-Mu\~noz, Pablo and
            Roy, Hrittik and Hauberg, S{\o}ren},
            journal={arXiv preprint arXiv:2405.17277},
            year={2024}
        }
        ```
    """
    reortho_expected = ["none", "full"]
    if reortho not in reortho_expected:
        msg = f"Unexpected input for {reortho}: either of {reortho_expected} expected."
        raise TypeError(msg)

    def estimate(matvec, v, *params):
        matvec_convert, aux_args = func.closure_convert(matvec, v, *params)
        return _estimate(matvec_convert, v, *params, *aux_args)

    def _estimate(matvec_convert: Callable, v, *params):
        reortho_ = reortho_vjp if reortho_vjp != "match" else reortho_vjp
        return _hessenberg_forward(
            matvec_convert, num_matvecs, v, *params, reortho=reortho_
        )

    def estimate_fwd(matvec_convert: Callable, v, *params):
        outputs = _estimate(matvec_convert, v, *params)
        return outputs, (outputs, params)

    def estimate_bwd(matvec_convert: Callable, cache, vjp_incoming):
        (Q, H, r, c), params = cache
        dQ, dH, dr, dc = vjp_incoming

        return _hessenberg_adjoint(
            matvec_convert,
            *params,
            Q=Q,
            H=H,
            r=r,
            c=c,
            dQ=dQ,
            dH=dH,
            dr=dr,
            dc=dc,
            reortho=reortho,
        )

    if custom_vjp:
        _estimate = func.custom_vjp(_estimate, nondiff_argnums=(0,))
        _estimate.defvjp(estimate_fwd, estimate_bwd)  # type: ignore
    return estimate


def _hessenberg_forward(matvec, num_matvecs, v, *params, reortho: str):
    if num_matvecs < 0 or num_matvecs > len(v):
        msg = _error_num_matvecs(num_matvecs, maxval=len(v), minval=0)
        raise ValueError(msg)

    # Initialise the variables
    (n,), k = np.shape(v), num_matvecs
    Q = np.zeros((n, k), dtype=v.dtype)
    H = np.zeros((k, k), dtype=v.dtype)
    initlength = np.sqrt(linalg.inner(v, v))
    init = (Q, H, v, initlength)

    if num_matvecs == 0:
        return _DecompResult(
            Q_tall=Q, J_small=H, residual=v, init_length_inv=1 / initlength
        )

    # Fix the step function
    def forward_step(i, val):
        return _hessenberg_forward_step(*val, matvec, *params, idx=i, reortho=reortho)

    # Loop and return
    Q, H, v, _length = control_flow.fori_loop(0, k, forward_step, init)
    return _DecompResult(
        Q_tall=Q, J_small=H, residual=v, init_length_inv=1 / initlength
    )


def _hessenberg_forward_step(Q, H, v, length, matvec, *params, idx, reortho: str):
    # Save
    v /= length
    Q = Q.at[:, idx].set(v)

    # Evaluate
    v = matvec(v, *params)

    # Orthonormalise
    h = Q.T @ v
    v = v - Q @ h

    # Re-orthonormalise
    if reortho != "none":
        v = v - Q @ (Q.T @ v)

    # Read the length
    length = np.sqrt(linalg.inner(v, v))

    # Save
    h = h.at[idx + 1].set(length)
    H = H.at[:, idx].set(h)

    return Q, H, v, length


def _hessenberg_adjoint(matvec, *params, Q, H, r, c, dQ, dH, dr, dc, reortho: str):
    # Extract the matrix shapes from Q
    _, num_matvecs = np.shape(Q)

    # Prepare a bunch of auxiliary matrices

    def lower(m):
        m_tril = np.tril(m)
        return m_tril - 0.5 * _extract_diag(m_tril)

    e_1, e_K = np.eye(num_matvecs)[[0, -1], :]
    lower_mask = lower(np.ones((num_matvecs, num_matvecs)))

    # Initialise
    eta = dH @ e_K - Q.T @ dr
    lambda_k = dr + Q @ eta
    Lambda = np.zeros_like(Q)
    Gamma = np.zeros_like(dQ.T @ Q)
    dp = tree.tree_map(np.zeros_like, params)

    # Prepare more  auxiliary matrices
    Pi_xi = dQ.T + linalg.outer(eta, r)
    Pi_gamma = -dc * c * linalg.outer(e_1, e_1) + H @ dH.T - (dQ.T @ Q)

    # Prepare reorthogonalisation:
    P = Q.T
    ps = dH.T
    ps_mask = np.tril(np.ones((num_matvecs, num_matvecs)), 1)

    # Loop over those values
    indices = np.arange(0, len(H), step=1)
    beta_minuses = np.concatenate([np.ones((1,)), linalg.diagonal(H, -1)])
    alphas = linalg.diagonal(H)
    beta_pluses = H - _extract_diag(H) - _extract_diag(H, -1)
    scan_over = {
        "beta_minus": beta_minuses,
        "alpha": alphas,
        "beta_plus": beta_pluses,
        "idx": indices,
        "lower_mask": lower_mask,
        "Pi_gamma": Pi_gamma,
        "Pi_xi": Pi_xi,
        "p": ps,
        "p_mask": ps_mask,
        "q": Q.T,
    }

    # Fix the step function
    def adjoint_step(x, y):
        output = _hessenberg_adjoint_step(
            *x, **y, matvec=matvec, params=params, Q=Q, reortho=reortho
        )
        return output, ()

    # Scan
    init = (lambda_k, Lambda, Gamma, P, dp)
    result, _ = control_flow.scan(adjoint_step, init, xs=scan_over, reverse=True)
    (lambda_k, Lambda, Gamma, _P, dp) = result

    # Solve for the input gradient
    dv = lambda_k * c

    return dv, *dp


def _hessenberg_adjoint_step(
    # Running variables
    lambda_k,
    Lambda,
    Gamma,
    P,
    dp,
    *,
    # Matrix-vector product
    matvec,
    params,
    # Loop over: index
    idx,
    # Loop over: submatrices of H
    beta_minus,
    alpha,
    beta_plus,
    # Loop over: auxiliary variables for Gamma
    lower_mask,
    Pi_gamma,
    Pi_xi,
    q,
    # Loop over: reorthogonalisation
    p,
    p_mask,
    # Other parameters
    Q,
    reortho: str,
):
    # Reorthogonalise
    if reortho == "full":
        P = p_mask[:, None] * P
        p = p_mask * p
        lambda_k = lambda_k - P.T @ (P @ lambda_k) + P.T @ p

    # Transposed matvec and parameter-gradient in a single matvec
    _, vjp = func.vjp(lambda u, v: matvec(u, *v), q, params)
    vecmat_lambda, dp_increment = vjp(lambda_k)
    dp = tree.tree_map(lambda g, h: g + h, dp, dp_increment)

    # Solve for (Gamma + Gamma.T) e_K
    tmp = lower_mask * (Pi_gamma - vecmat_lambda @ Q)
    Gamma = Gamma.at[idx, :].set(tmp)

    # Solve for the next lambda (backward substitution step)
    Lambda = Lambda.at[:, idx].set(lambda_k)
    xi = Pi_xi + (Gamma + Gamma.T)[idx, :] @ Q.T
    lambda_k = xi - (alpha * lambda_k - vecmat_lambda) - beta_plus @ Lambda.T
    lambda_k /= beta_minus
    return lambda_k, Lambda, Gamma, P, dp


def _extract_diag(x, offset=0):
    diag = linalg.diagonal(x, offset=offset)
    return linalg.diagonal_matrix(diag, offset=offset)


def bidiag(num_matvecs: int, /, materialize: bool = True):
    """Construct an implementation of **bidiagonalisation**.

    Uses pre-allocation and full reorthogonalisation.

    Works for **arbitrary matrices**. No symmetry required.

    Decompose a matrix into a product of orthogonal-**bidiagonal**-orthogonal matrices.
    Use this algorithm for approximate **singular value** decompositions.

    Internally, Matfree uses JAX to turn matrix-vector- into vector-matrix-products.

    ??? note "A note about differentiability"
        Unlike [tridiag_sym][matfree.decomp.tridiag_sym] or
        [hessenberg][matfree.decomp.hessenberg], this function's reverse-mode
        derivatives are very efficient. Custom gradients for bidiagonalisation
        are a work in progress, and if you need to differentiate the decompositions,
        consider using [tridiag_sym][matfree.decomp.tridiag_sym] for the time being.

    """

    def estimate(Av: Callable, v0, *parameters):
        # Infer the size of A from v0
        (ncols,) = np.shape(v0)
        w0_like = func.eval_shape(Av, v0, *parameters)
        (nrows,) = np.shape(w0_like)

        # Complain if the shapes don't match
        max_num_matvecs = min(nrows, ncols)
        if num_matvecs > max_num_matvecs or num_matvecs < 0:
            msg = _error_num_matvecs(num_matvecs, maxval=min(nrows, ncols), minval=0)
            raise ValueError(msg)

        v0_norm, length = _normalise(v0)
        init_val = init(v0_norm, nrows=nrows, ncols=ncols)

        if num_matvecs == 0:
            uk_all_T, J, vk_all, (beta, vk) = extract(init_val)
            return _DecompResult(
                Q_tall=(uk_all_T, vk_all.T),
                J_small=J,
                residual=beta * vk,
                init_length_inv=1 / length,
            )

        def body_fun(_, s):
            return step(Av, s, *parameters)

        result = control_flow.fori_loop(
            0, num_matvecs, body_fun=body_fun, init_val=init_val
        )
        uk_all_T, J, vk_all, (beta, vk) = extract(result)
        return _DecompResult(
            Q_tall=(uk_all_T, vk_all.T),
            J_small=J,
            residual=beta * vk,
            init_length_inv=1 / length,
        )

    class State(containers.NamedTuple):
        i: int
        Us: Array
        Vs: Array
        alphas: Array
        betas: Array
        beta: Array
        vk: Array

    def init(init_vec: Array, *, nrows, ncols) -> State:
        alphas = np.zeros((num_matvecs,))
        betas = np.zeros((num_matvecs,))
        Us = np.zeros((num_matvecs, nrows))
        Vs = np.zeros((num_matvecs, ncols))
        v0, _ = _normalise(init_vec)
        return State(0, Us, Vs, alphas, betas, np.zeros(()), v0)

    def step(Av, state: State, *parameters) -> State:
        i, Us, Vs, alphas, betas, beta, vk = state
        Vs = Vs.at[i].set(vk)
        betas = betas.at[i].set(beta)

        # Use jax.vjp to evaluate the vector-matrix product
        Av_eval, vA = func.vjp(lambda v: Av(v, *parameters), vk)
        uk = Av_eval - beta * Us[i - 1]
        uk, alpha = _normalise(uk)
        uk, *_ = _gram_schmidt_classical(uk, Us)  # full reorthogonalisation
        Us = Us.at[i].set(uk)
        alphas = alphas.at[i].set(alpha)

        (vA_eval,) = vA(uk)
        vk = vA_eval - alpha * vk
        vk, beta = _normalise(vk)
        vk, *_ = _gram_schmidt_classical(vk, Vs)  # full reorthogonalisation

        return State(i + 1, Us, Vs, alphas, betas, beta, vk)

    def extract(state: State, /):
        _, uk_all, vk_all, alphas, betas, beta, vk = state

        if materialize:
            B = _todense_bidiag(alphas, betas[1:])
            return uk_all.T, B, vk_all, (beta, vk)

        return uk_all.T, (alphas, betas[1:]), vk_all, (beta, vk)

    def _gram_schmidt_classical(vec, vectors):  # Gram-Schmidt
        vec, coeffs = control_flow.scan(_gram_schmidt_classical_step, vec, xs=vectors)
        vec, length = _normalise(vec)
        return vec, length, coeffs

    def _gram_schmidt_classical_step(vec1, vec2):
        coeff = linalg.inner(vec1, vec2)
        vec_ortho = vec1 - coeff * vec2
        return vec_ortho, coeff

    def _normalise(vec):
        length = linalg.vector_norm(vec)
        return vec / length, length

    def _todense_bidiag(d, e):
        diag = linalg.diagonal_matrix(d)
        offdiag = linalg.diagonal_matrix(e, 1)
        return diag + offdiag

    return estimate


def _error_num_matvecs(num, maxval, minval):
    msg1 = f"Parameter 'num_matvecs'={num} exceeds the acceptable range. "
    msg2 = f"Expected: {minval} <= num_matvecs <= {maxval}."
    return msg1 + msg2



"""Carry out stochastic trace estimation with minimal memory.

Matfree's implementation of stochastic trace estimation
via Hutchinson's method defaults to computing all
Monte-Carlo samples at once, because this is the fastest
implementation as long as all samples fit into memory.


Some matrix-vector products, however, are so large that
we can only store a single sample in memory at once.
Here is how we can wrap calls around the trace estimators
in such a scenario to save memory.
"""

import functools

import jax
import jax.numpy as jnp

from matfree import stochtrace

# ## Stochastic trace estimation
#
# The conventional setup for estimating the trace of a large matrix
# would look like this.

nrows = 100  # but imagine nrows=100,000,000,000 instead
nsamples = 1_000


def large_matvec(v):
    """Evaluate a (dummy for a) large matrix-vector product."""
    return 1.2345 * v


integrand = stochtrace.integrand_trace()
x0 = jnp.ones((nrows,))
sampler = stochtrace.sampler_rademacher(x0, num=nsamples)
estimate = stochtrace.estimator(integrand, sampler)

key = jax.random.PRNGKey(1)
trace = estimate(large_matvec, key)
print(trace)


# The above code requires nrows $\times$ nsamples storage, which
# is prohibitive for extremely large matrices.
# Instead, we can loop around estimate() to do the following:
# The below code requires nrows $\times$ 1 storage:

sampler = stochtrace.sampler_rademacher(x0, num=1)
estimate = stochtrace.estimator(integrand, sampler)
estimate = functools.partial(estimate, large_matvec)

key = jax.random.PRNGKey(2)
keys = jax.random.split(key, num=nsamples)
traces = jax.lax.map(estimate, keys)
trace = jnp.mean(traces)
print(trace)

# In practice, we often combine both approaches by choosing
# the largest nsamples (in the first implementation) so that
# nrows $\times$ nsamples fits into memory, and handle all samples beyond
# that via the split-and-map combination.
#
#
# If we reverse-mode differentiate through the sampler, we have to
# be careful because by default, reverse-mode differentiation
# stores all intermediate results (and the memory-efficiency of using
# jax.lax.map is void).
# To solve this problem, place a jax.checkpoint around the estimator:

traces = jax.lax.map(jax.checkpoint(estimate), keys)
trace = jnp.mean(traces)
print(trace)

# This implementation recomputes the forward pass for each key during the
# backward pass, but preserves the memory-efficiency on the backward pass.
#
#
# In summary, memory efficiency can be achieved by calling estimators
# inside jax.lax.map (with or without checkpoints).



"""Implement uncertainty quantification for trace estimation.

Computing higher moments of trace-estimates can easily
be turned into uncertainty quantification.
"""

import jax
import jax.numpy as jnp

from matfree import stochtrace

A = jnp.reshape(jnp.arange(36.0), (6, 6)) / 36


def matvec(x):
    """Evaluate a matrix-vector product."""
    return A.T @ (A @ x) + x


x_like = jnp.ones((6,))
num_samples = 10_000

# ## Higher moments
#
# Trace estimation involves estimating expected values of random variables.
# Sometimes, second and higher moments of a random variable are interesting.

normal = stochtrace.sampler_normal(x_like, num=num_samples)
integrand = stochtrace.integrand_trace()
integrand = stochtrace.integrand_wrap_moments(integrand, [1, 2])
estimator = stochtrace.estimator(integrand, sampler=normal)
first, second = estimator(matvec, jax.random.PRNGKey(1))


# For normally-distributed base-samples,
# we know that the variance is twice the squared Frobenius norm.


print(second - first**2)
print(2 * jnp.linalg.norm(A.T @ A + jnp.eye(6), ord="fro") ** 2)


# ## Uncertainty quantification
#
# Variance estimation leads to uncertainty quantification:
# The variance of the estimator is equal to the variance of the random variable
# divided by the number of samples.

variance = (second - first**2) / num_samples
print(variance)



"""Test stochastic Lanczos quadrature for Schatten-p-norms."""

from matfree import decomp, funm, stochtrace, test_util
from matfree.backend import linalg, np, prng, testing


def make_A(nrows, ncols, num_significant_singular_vals):
    """Make a positive definite matrix with certain spectrum."""
    # 'Invent' a spectrum. Use the number of pre-defined eigenvalues.
    n = min(nrows, ncols)
    d = np.arange(n) + 1.0
    d = d.at[num_significant_singular_vals:].set(0.001)
    return test_util.asymmetric_matrix_from_singular_values(d, nrows=nrows, ncols=ncols)


@testing.parametrize("nrows", [30])
@testing.parametrize("ncols", [30])
@testing.parametrize("num_significant_singular_vals", [30])
@testing.parametrize("num_matvecs", [20])
@testing.parametrize("power", [1, 2, 5])
def test_schatten_norm(nrows, ncols, num_significant_singular_vals, num_matvecs, power):
    """Assert that the Schatten norm is accurate."""
    A = make_A(nrows, ncols, num_significant_singular_vals)
    _, s, _ = linalg.svd(A, full_matrices=False)
    expected = np.sum(s**power)

    _, ncols = np.shape(A)
    args_like = np.ones((ncols,), dtype=float)
    sampler = stochtrace.sampler_normal(args_like, num=500)
    bidiag = decomp.bidiag(num_matvecs)
    integrand = funm.integrand_funm_product_schatten_norm(power, bidiag)
    estimate = stochtrace.estimator(integrand, sampler)

    key = prng.prng_key(1)
    received = estimate(lambda v: A @ v, key)

    print_if_assert_fails = ("error", np.abs(received - expected), "target:", expected)
    assert np.allclose(received, expected, atol=1e-2, rtol=1e-2), print_if_assert_fails



"""Test matrix-function-vector products via Lanczos' algorithm."""

from matfree import decomp, funm, test_util
from matfree.backend import linalg, np, prng, testing


@testing.parametrize("dense_funm", [funm.dense_funm_sym_eigh, funm.dense_funm_schur])
@testing.parametrize("reortho", ["full", "none"])
def test_funm_lanczos_sym_matches_eigh_implementation(dense_funm, reortho, n=11):
    """Test matrix-function-vector products via Lanczos' algorithm."""
    # Create a test-problem: matvec, matrix function,
    # vector, and parameters (a matrix).

    def matvec(x, p):
        return p @ x

    def fun(x):
        return np.sin(x)

    v = prng.normal(prng.prng_key(2), shape=(n,))

    eigvals = np.linspace(0.01, 0.99, num=n)
    matrix = test_util.symmetric_matrix_from_eigenvalues(eigvals)

    # Compute the solution
    eigvals, eigvecs = linalg.eigh(matrix)
    log_matrix = eigvecs @ linalg.diagonal(fun(eigvals)) @ eigvecs.T
    expected = log_matrix @ v

    # Compute the matrix-function vector product
    dense_funm = dense_funm(fun)
    lanczos = decomp.tridiag_sym(6, materialize=True, reortho=reortho)
    matfun_vec = funm.funm_lanczos_sym(dense_funm, lanczos)
    received = matfun_vec(matvec, v, matrix)
    assert np.allclose(expected, received, atol=1e-6)



"""Tests for Bai and Golub's log-determinant bounds."""

from matfree import bounds, test_util
from matfree.backend import linalg, np


def test_logdet():
    """Test that Bai and Golub's log-determinant bound is correct."""
    # Set up a test-problem.
    eigvals = np.asarray([1.0, 2.0, 3.0, 4.0])
    matrix = test_util.symmetric_matrix_from_eigenvalues(eigvals)

    # Compute the bound
    trace = linalg.trace(matrix)
    fronorm = linalg.matrix_norm(matrix, which="fro") ** 2
    eigvals_bounds = (0.9, 4.1)
    lower = bounds.baigolub96_logdet_spd(eigvals_bounds[0], 4, trace, fronorm)
    upper = bounds.baigolub96_logdet_spd(eigvals_bounds[1], 4, trace, fronorm)

    # Reference
    _sign, logdet = linalg.slogdet(matrix)

    # Assert that the bounds are satisfied,
    # but also that they are somewhat tight
    # (because we use tight eigenvalue bounds)
    assert 0.9 * logdet < lower < logdet < upper < 1.1 * logdet



"""Types."""

# fmt: off
from collections.abc import Callable  # noqa: F401
from typing import (  # noqa: F401, UP035
    Any,
    Generic,
    Iterable,
    Sequence,
    Tuple,
    TypeVar,
    Union,
)

from jax import Array  # noqa: F401

# fmt: on



"""Combine trace estimation with control variates.

Here is how to implement control variates.
"""

import jax
import jax.numpy as jnp

from matfree import stochtrace

# Create a matrix to whose trace/diagonal to approximate.


nrows, ncols = 4, 4
A = jnp.reshape(jnp.arange(1.0, 1.0 + nrows * ncols), (nrows, ncols))


# Set up the sampler.

x_like = jnp.ones((ncols,), dtype=float)
sample_fun = stochtrace.sampler_normal(x_like, num=10_000)


# First, compute the diagonal.


problem = stochtrace.integrand_diagonal()
estimate = stochtrace.estimator(problem, sample_fun)
diagonal_ctrl = estimate(lambda v: A @ v, jax.random.PRNGKey(1))


# Then, compute trace and diagonal jointly
# using the estimate of the diagonal as a control variate.


def matvec_ctrl(v):
    """Evaluate a matrix-vector product with a control variate."""
    return A @ v - diagonal_ctrl * v


problem = stochtrace.integrand_trace_and_diagonal()
estimate = stochtrace.estimator(problem, sample_fun)
trace_and_diagonal = estimate(matvec_ctrl, jax.random.PRNGKey(2))
trace, diagonal = trace_and_diagonal["trace"], trace_and_diagonal["diagonal"]


# We can, of course, compute it without a control variate as well.

problem = stochtrace.integrand_trace_and_diagonal()
estimate = stochtrace.estimator(problem, sample_fun)
trace_and_diagonal = estimate(lambda v: A @ v, jax.random.PRNGKey(2))
trace_ref, diagonal_ref = trace_and_diagonal["trace"], trace_and_diagonal["diagonal"]


# Compare the results.
# First, the diagonal.


print("True value:", jnp.diag(A))
print("Control variate:", diagonal_ctrl, jnp.linalg.norm(jnp.diag(A) - diagonal_ctrl))
print("Approximation:", diagonal_ref, jnp.linalg.norm(jnp.diag(A) - diagonal_ref))
print(
    "Control-variate approximation:",
    diagonal + diagonal_ctrl,
    jnp.linalg.norm(jnp.diag(A) - diagonal - diagonal_ctrl),
)


# Then, the trace.


print("True value:", jnp.trace(A))
print(
    "Control variate:",
    jnp.sum(diagonal_ctrl),
    jnp.abs(jnp.trace(A) - jnp.sum(diagonal_ctrl)),
)
print("Approximation:", trace_ref, jnp.abs(jnp.trace(A) - trace_ref))
print(
    "Control variate approximation:",
    trace + jnp.sum(diagonal_ctrl),
    jnp.abs(jnp.trace(A) - trace - jnp.sum(diagonal_ctrl)),
)



"""Create developer documentation from the README."""

import contextlib
import os
from re import sub


def mkdir_unless_exists(f: str, /) -> None:
    """Create a directory."""
    with contextlib.suppress(FileExistsError):
        os.mkdir(f)


def filename(s, remove):
    """Turn a string into a file-name.

    Removes a list of sub-strings and returns a snake_case version.
    """
    for r in remove:
        s = s.replace(r, "")
    return snake_case(s)


def snake_case(s):
    """Convert a string to snake case.

    Source:
    https://www.w3resource.com/python-exercises/string/python-data-type-string-exercise-97.php
    """
    # Replace hyphens with spaces,
    # then apply regular expression substitutions for title case conversion
    # and add an underscore between words,
    # finally convert the result to lowercase
    return "_".join(
        sub(
            "([A-Z][a-z]+)", r" \1", sub("([A-Z]+)", r" \1", s.replace("-", " "))
        ).split()
    ).lower()


def h2_to_h1(string):
    """Turn a 2-header into a 1-header.

    Remove initial '#'.
    """
    return string[1:]


if __name__ == "__main__":
    # Create dev docs
    source = "README.md"
    directory = "docs/Developer_documentation/"
    mkdir_unless_exists("docs/Developer_documentation/")

    # Read the README
    with open(source) as file:
        full_readme = file.readlines()

    # Find locations of H2's
    is_header = [x[:3] == "## " for x in full_readme]
    indices = [i for i, x in enumerate(is_header) if x]

    # Split the README into "intro" and dev_docs

    # First block is the index
    block = full_readme[: indices[0]]
    title = filename(block[0], remove=["# ", ":"])

    # Save index as file
    with open("docs/index.md", "w") as file:
        file.writelines(block)

    # Create all dev-docs pages

    indices = [*indices, -1]
    for i, (idx_from, idx_to) in enumerate(zip(indices[:-1], indices[1:])):
        # Extract subsection of README
        block = full_readme[idx_from:idx_to]

        # Create a title
        block[0] = h2_to_h1(block[0])
        string = filename(block[0], remove=["# ", ":"])
        title = f"{1+i}_{string}"

        # Save as file
        with open(f"{directory}{title}.md", "w") as file:
            file.writelines(block)



"""Test the estimation of squared Frobenius-norms."""

from matfree import stochtrace
from matfree.backend import func, linalg, np, prng, tree


def test_frobeniusnorm_squared():
    """Assert that the estimated squared Frobenius norm approximates accurately."""

    def fun(x):
        """Create a nonlinear, to-be-differentiated function."""
        fx = np.sin(np.flip(np.cos(x["params"])) + 1.0) * np.sin(x["params"])
        return {"params": fx}

    key = prng.prng_key(seed=2)

    # Linearise function
    x0 = prng.uniform(key, shape=(4,))  # random lin. point
    args_like = {"params": x0}
    _, jvp = func.linearize(fun, args_like)
    [J] = tree.tree_leaves(func.jacfwd(fun)(args_like))
    expected = linalg.trace(J.T @ J)

    # Estimate the matrix function
    problem = stochtrace.integrand_frobeniusnorm_squared()
    sampler = stochtrace.sampler_rademacher(args_like, num=100_000)
    estimate = stochtrace.estimator(problem, sampler=sampler)
    received = estimate(jvp, key)

    assert np.allclose(expected, received, rtol=1e-2)



"""PyTree utilities."""

import jax
import jax.flatten_util


def tree_all(pytree, /):
    return jax.tree.all(pytree)


def tree_map(func, pytree, *rest):
    return jax.tree.map(func, pytree, *rest)


def ravel_pytree(pytree, /):
    ravelled, unravel = jax.flatten_util.ravel_pytree(pytree)

    # Wrap through a partial_pytree() to make ravel_pytree
    # compatible with jax.eval_shape
    return ravelled, partial_pytree(unravel)


def tree_leaves(pytree, /):
    return jax.tree.leaves(pytree)


def tree_structure(pytree, /):
    return jax.tree.structure(pytree)


def partial_pytree(func, /):
    return jax.tree_util.Partial(func)



"""Low-rank approximations (like partial Cholesky decompositions) of matrices."""

from matfree.backend import control_flow, func, linalg, np
from matfree.backend.typing import Array, Callable


def preconditioner(cholesky: Callable, /) -> Callable:
    r"""Turn a low-rank approximation into a preconditioner.

    Parameters
    ----------
    cholesky
        (Partial) Cholesky decomposition.
        Usually, the result of either
        [cholesky_partial][matfree.low_rank.cholesky_partial]
        or
        [cholesky_partial_pivot][matfree.low_rank.cholesky_partial_pivot].


    Returns
    -------
    solve
        A function that computes

        $$
        (v, s, *p) \mapsto (sI + L(*p) L(*p)^\top)^{-1} v,
        $$

        where $K = [k(i,j,*p)]_{ij} \approx L(*p) L(*p)^\top$
        and $L$ comes from the low-rank approximation.
    """

    def solve(v: Array, s: float, *cholesky_params):
        chol, info = cholesky(*cholesky_params)

        # Assert that the low-rank matrix is tall,
        # not wide (every sign has a story...)
        N, n = np.shape(chol)
        assert n <= N, (N, n)

        # Scale
        U = chol / np.sqrt(s)
        V = chol.T / np.sqrt(s)
        v /= s

        # Cholesky decompose the capacitance matrix
        # and solve the system
        eye_n = np.eye(n)
        chol_cap = linalg.cho_factor(eye_n + V @ U)
        sol = linalg.cho_solve(chol_cap, V @ v)
        return v - U @ sol, info

    return solve


def cholesky_partial(mat_el: Callable, /, *, nrows: int, rank: int) -> Callable:
    """Compute a partial Cholesky factorisation."""

    def cholesky(*params):
        if rank > nrows:
            msg = f"Rank exceeds n: {rank} >= {nrows}."
            raise ValueError(msg)
        if rank < 1:
            msg = f"Rank must be positive, but {rank} < {1}."
            raise ValueError(msg)

        step = _cholesky_partial_body(mat_el, nrows, *params)
        chol = np.zeros((nrows, rank))
        return control_flow.fori_loop(0, rank, step, chol), {}

    return cholesky


def _cholesky_partial_body(fn: Callable, n: int, *args):
    idx = np.arange(n)

    def matrix_element(i, j):
        return fn(i, j, *args)

    def matrix_column(i):
        fun = func.vmap(matrix_element, in_axes=(0, None))
        return fun(idx, i)

    def body(i, L):
        element = matrix_element(i, i)
        l_ii = np.sqrt(element - linalg.inner(L[i], L[i]))

        column = matrix_column(i)
        l_ji = column - L @ L[i, :]
        l_ji /= l_ii

        return L.at[:, i].set(l_ji)

    return body


def cholesky_partial_pivot(mat_el: Callable, /, *, nrows: int, rank: int) -> Callable:
    """Compute a partial Cholesky factorisation with pivoting."""

    def cholesky(*params):
        if rank > nrows:
            msg = f"Rank exceeds nrows: {rank} >= {nrows}."
            raise ValueError(msg)
        if rank < 1:
            msg = f"Rank must be positive, but {rank} < {1}."
            raise ValueError(msg)

        body = _cholesky_partial_pivot_body(mat_el, nrows, *params)

        L = np.zeros((nrows, rank))
        P = np.arange(nrows)

        init = (L, P, P, True)
        (L, P, _matrix, success) = control_flow.fori_loop(0, rank, body, init)
        return _pivot_invert(L, P), {"success": success}

    return cholesky


def _cholesky_partial_pivot_body(fn: Callable, n: int, *args):
    idx = np.arange(n)

    def matrix_element(i, j):
        return fn(i, j, *args)

    def matrix_element_p(i, j, *, permute):
        return matrix_element(permute[i], permute[j])

    def matrix_column_p(i, *, permute):
        fun = func.vmap(matrix_element, in_axes=(0, None))
        return fun(permute[idx], permute[i])

    def matrix_diagonal_p(*, permute):
        fun = func.vmap(matrix_element)
        return fun(permute[idx], permute[idx])

    def body(i, carry):
        L, P, P_matrix, success = carry

        # Access the matrix
        diagonal = matrix_diagonal_p(permute=P_matrix)

        # Find the largest entry for the residuals
        residual_diag = diagonal - func.vmap(linalg.inner)(L, L)
        res = np.abs(residual_diag)
        k = np.argmax(res)

        # Pivot [pivot!!! pivot!!! pivot!!! :)]
        P_matrix = _swap_cols(P_matrix, i, k)
        L = _swap_rows(L, i, k)
        P = _swap_rows(P, i, k)

        # Access the matrix
        element = matrix_element_p(i, i, permute=P_matrix)
        column = matrix_column_p(i, permute=P_matrix)

        # Perform a Cholesky step
        # (The first line could also be accessed via
        #  residual_diag[k], but it might
        #  be more readable to do it again)
        l_ii_squared = element - linalg.inner(L[i], L[i])
        l_ii = np.sqrt(l_ii_squared)
        l_ji = column - L @ L[i, :]
        l_ji /= l_ii
        success = np.logical_and(success, l_ii_squared > 0.0)

        # Update the estimate
        L = L.at[:, i].set(l_ji)
        return L, P, P_matrix, success

    return body


def _swap_cols(arr, i, j):
    return _swap_rows(arr.T, i, j).T


def _swap_rows(arr, i, j):
    ai, aj = arr[i], arr[j]
    arr = arr.at[i].set(aj)
    return arr.at[j].set(ai)


def _pivot_invert(arr, pivot, /):
    """Invert and apply a pivoting array to a matrix."""
    return arr[np.argsort(pivot)]



"""Ensure that bidiag, tridiag, etc. have consistent signatures."""

from matfree import decomp
from matfree.backend import np, prng, testing, tree


def case_method_bidiag():
    return decomp.bidiag


def case_method_tridiag_sym():
    return lambda n: decomp.tridiag_sym(n, reortho="none")


def case_method_tridiag_sym_reortho():
    return lambda n: decomp.tridiag_sym(n, reortho="full")


def case_method_hessenberg():
    return lambda n: decomp.hessenberg(n, reortho="none")


def case_method_hessenberg_reortho():
    return lambda n: decomp.hessenberg(n, reortho="full")


@testing.parametrize("nrows", [13])
@testing.parametrize("num_matvecs", [6, 13, 0])
@testing.parametrize_with_cases("method", cases=".", prefix="case_method_")
def test_output_shape_as_expected(nrows, num_matvecs, method):
    """Test that all factorisation methods yield consistent output shapes."""
    key = prng.prng_key(1)
    key1, key2 = prng.split(key, num=2)
    A = prng.normal(key1, shape=(nrows, nrows))
    v0 = prng.normal(key2, shape=(nrows,))

    algorithm = method(num_matvecs)
    Us, B, res, ln = algorithm(lambda v: A @ v, v0)

    # Normalise the Us to always have a list
    Us = tree.tree_leaves(Us)

    for U in Us:
        assert np.shape(U) == (nrows, num_matvecs)

    assert np.shape(B) == (num_matvecs, num_matvecs)
    assert np.shape(res) == (nrows,)
    assert np.shape(ln) == ()


@testing.parametrize("nrows", [13])
@testing.parametrize("num_matvecs", [-1, 14])  # 0 and 13 must work (see above)
@testing.parametrize_with_cases("method", cases=".", prefix="case_method_")
def test_value_error_for_unusual_num_matvecs(nrows, num_matvecs, method):
    """Assert a ValueError is raised when the num_matvecs exceeds the matrix size."""
    key = prng.prng_key(1)
    key1, key2 = prng.split(key, num=2)
    A = prng.normal(key1, shape=(nrows, nrows))
    v0 = prng.normal(key2, shape=(nrows,))

    with testing.raises(ValueError, match="exceeds"):
        alg = method(num_matvecs)
        _ = alg(lambda v, p: p @ v, v0, A)
